
/*****************************************************************************
** Function: AddHistory()
*/

static void AddHistory(struct Text *Self, STRING History)
{
   LONG i;

   /*** Push all the numbers up by one ***/
   for (i=0; i < Self->HistorySize; i++) {
      if (Self->History[i].Number > 0) Self->History[i].Number++;
      if (Self->History[i].Number > Self->HistorySize) Self->History[i].Number = 0;
   }

   /*** Find an empty buffer and add the history text to it ***/
   for (i=0; i < Self->HistorySize; i++) {
      if (Self->History[i].Number IS 0) {
         Self->History[i].Number = 1; /* New entries have a number of 1 */
         StrCopy(History, Self->History[i].Buffer, sizeof(Self->History[0].Buffer));
         return;
      }
   }
}

/*****************************************************************************
** Internal: AddLine()
*/

static ERROR AddLine(struct Text *Self, STRING String, LONG Line, LONG Length, LONG Allocated)
{
   struct TextLine *newlines;
   STRING str;
   LONG len, i, line, unicodelen;

   /* If a line number was not given then we will insert the line at the
   ** end of the list (note that the first line starts at #1).
   */

   if (Line < 0) line = Self->AmtLines;
   else line = Line;

   /*** Get the length of the text ***/

   if ((String) AND (*String)) {
      if (Length >= 0) len = Length;
      else for (len=0; (String[len]) AND (String[len] != '\n') AND (String[len] != '\r'); len++);
   }
   else len = 0;

   /*** Stop the string from exceeding the acceptable character limit ***/

   if (len >= Self->CharLimit) {
      for (unicodelen=0, i=0; (i < len) AND (unicodelen < Self->CharLimit); unicodelen++) {
         for (++i; (String[i] & 0xc0) IS 0x80; i++);
      }
      len = i;
   }

   /*** If the line array is at capacity, expand it ***/

   if (Self->AmtLines >= Self->MaxLines) {
      if (AllocMemory(sizeof(struct TextLine) * (Self->MaxLines + 100), MEM_DATA, (APTR *)&newlines, NULL) IS ERR_Okay) {
         Self->MaxLines += 100;
         if (Self->Array) {
            CopyMemory(Self->Array, newlines, sizeof(Self->Array[0]) * Self->AmtLines);
            FreeMemory(Self->Array);
         }
         Self->Array = newlines;
      }
      else return(ERR_AllocMemory);
   }

   /*** Expand the end of the array to allow for our new entry ***/

   if (line < Self->AmtLines) {
      CopyMemory(Self->Array + line, Self->Array + line + 1, sizeof(Self->Array[0]) * (Self->AmtLines - line));
   }

   Self->AmtLines++;

   /*** Insert the new line into the array ***/

   if (len > 0) {
      if (Allocated IS TRUE) {
         Self->Array[line].String      = String;
         Self->Array[line].Tags        = NULL;
         Self->Array[line].Length      = len;
         Self->Array[line].PixelLength = CalculateWidth(Self, String, len);
      }
      else if (AllocMemory(len + 1, MEM_STRING|MEM_NOCLEAR, (APTR *)&str, NULL) IS ERR_Okay) {
         for (i=0; (i < len) AND (String[i]); i++) str[i] = String[i];
         str[i] = 0;
         Self->Array[line].String      = str;
         Self->Array[line].Tags        = NULL;
         Self->Array[line].Length      = len;
         Self->Array[line].PixelLength = CalculateWidth(Self, str, len);
      }
      else {
         ClearMemory(Self->Array + line, sizeof(Self->Array[line]));
         return(ERR_AllocMemory);
      }
   }
   else ClearMemory(Self->Array + line, sizeof(Self->Array[line]));

   if (!Self->NoUpdate) {
      CalculateHScroll(Self);
      CalculateVScroll(Self);

      DrawLines(Self, line, Self->AmtLines - line);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: AddXML()
*/

static ERROR AddXML(struct Text *Self, struct XMLTag *XMLList, WORD Preformatted, LONG Line)
{
   struct TextLine *newlines;
   struct XMLTag *tag;
   LONG len, j, i;

   if ((!Self) OR (!XMLList)) return(ObjectError(Self, ERH_Function, ERR_Args));

   if (Line < 0) Line = Self->AmtLines;

   /*** Count the amount of bytes in the XML statement's content ***/

   len = 0;
   tag = XMLList->Child;
   while (tag) {
      len += XMLContentLength(tag);
      tag = tag->Next;
   }

   if (len > 0) {
      UBYTE str[len+1];

      /*** Copy the content into a string buffer ***/

      len = 0;
      tag = XMLList->Child;
      while (tag) {
         XMLExtractContent(tag, str, &len, Preformatted);
         tag = tag->Next;
      }

      /*** Add the line ***/

      if (Preformatted IS TRUE) {
         for (j=0; str[j];) {
            AddLine(Self, str+j, -1, -1, FALSE);
            while ((str[j]) AND (str[j] != '\n') AND (str[j] != '\r')) j++;
            while (str[j] IS '\r') j++;
            if (str[j] IS '\n') j++;
            while (str[j] IS '\r') j++;
         }
      }
      else {
         /*** Replace all white-space with real spaces (code 0x20) ***/

         for (j=0; j < len; j++) if (str[j] < 0x20) str[j] = ' ';

         /*** Shrink the string in areas where white-space is doubled-up ***/

         for (j=0; j < len; j++) {
            if ((str[j] IS '.') AND (str[j+1] IS 0x20) AND (str[j+2] IS 0x20)) {
               j += 2;  /* The end of sentences are allowed double-spaces ***/
            }
            else if ((str[j] IS 0x20) AND (str[j+1] IS 0x20)) {
               for (i=j; str[i] IS 0x20; i++);
               StrCopy(str+i, str+j, COPY_ALL);
            }
         }

         /*** Get the length of the text ***/

         for (len=0; str[len]; len++);

         /*** If the line array is at capacity, expand it ***/

         if (Self->AmtLines >= Self->MaxLines) {
            if (AllocMemory(sizeof(struct TextLine) * (Self->MaxLines + 100), MEM_DATA, (APTR *)&newlines, NULL) IS ERR_Okay) {
               Self->MaxLines += 100;
               if (Self->Array) {
                  CopyMemory(Self->Array, newlines, sizeof(Self->Array[0]) * Self->AmtLines);
                  FreeMemory(Self->Array);
               }
               Self->Array = newlines;
            }
            else return(ERR_AllocMemory);
         }

         /*** Expand the end of the array to allow for our new entry ***/

         if (Line < Self->AmtLines) {
            CopyMemory(Self->Array + Line, Self->Array + Line + 1, sizeof(Self->Array[0]) * (Self->AmtLines - Line));
         }

         Self->AmtLines++;

         /*** Insert the new line into the array ***/

         if (len > 0) {
            if (AllocMemory(len + 1, MEM_STRING|MEM_NOCLEAR, (APTR *)&Self->Array[Line].String, NULL) IS ERR_Okay) {
               Self->Array[Line].String[CharCopy(str, Self->Array[Line].String, len)] = 0;
               Self->Array[Line].Tags        = NULL;
               Self->Array[Line].Length      = len;
               Self->Array[Line].PixelLength = CalculateWidth(Self, str, len);
            }
            else {
               ClearMemory(Self->Array + Line, sizeof(Self->Array[Line]));
               return(ERR_AllocMemory);
            }
         }
         else ClearMemory(Self->Array + Line, sizeof(Self->Array[Line]));

         tag = XMLList->Child;
         while (tag) {
            if (StrCompare("b", tag->Attributes[0].Name, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
               InsertLineTag(Self, TF_BOLD, Line, NULL);
            }
            else if (StrCompare("i", tag->Attributes[0].Name, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
               InsertLineTag(Self, TF_ITALICS, Line, NULL);
            }
            else if (StrCompare("u", tag->Attributes[0].Name, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
               InsertLineTag(Self, TF_UNDERLINE, Line, NULL);
            }
            else if (StrCompare("font", tag->Attributes[0].Name, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
               for (j=1; j < tag->TotalAttributes; j++) {
                  if (StrCompare("colour", tag->Attributes[j].Name, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                     InsertLineTag(Self, TF_COLOUR, Line, tag->Attributes[j].Value);
                  }
               }
            }

            if (tag->Child) tag = tag->Child;
            else if (tag->Next) tag = tag->Next;
            else tag = NULL;
         }

         if (!Self->NoUpdate) {
            CalculateHScroll(Self);
            CalculateVScroll(Self);
            RedrawLine(Self, Line);
         }
      }
   }
   else AddLine(Self, "", -1, -1, FALSE);

   return(ERR_Okay);
}

/****************************************************************************/

static LONG XMLContentLength(struct XMLTag *XMLTag)
{
   LONG len;

   if (!XMLTag->Attributes[0].Name) {
      for (len=0; XMLTag->Attributes[0].Value[len]; len++);
   }
   else if ((XMLTag = XMLTag->Child)) {
      len = 0;
      while (XMLTag) {
         len += XMLContentLength(XMLTag);
         XMLTag = XMLTag->Next;
      }
   }
   else return(0);

   return(len);
}

/****************************************************************************/

static void XMLExtractContent(struct XMLTag *XMLTag, UBYTE *Buffer, LONG *Index, WORD Preformatted)
{
   STRING content;
   LONG i, pos;

   if (!XMLTag->Attributes[0].Name) {
      pos = *Index;
      if ((content = XMLTag->Attributes[0].Value)) {
         if (Preformatted IS TRUE) {
            for (i=0; content[i]; i++) Buffer[pos++] = content[i];
         }
         else for (i=0; content[i]; i++) {
            /*** Skip whitespace ***/
            if (content[i] <= 0x20) while ((content[i+1]) AND (content[i+1] <= 0x20)) i++;
            Buffer[pos++] = content[i];
         }
      }
      Buffer[pos] = 0;
      *Index = pos;
   }
   else if ((XMLTag = XMLTag->Child)) {
      while (XMLTag) {
         XMLExtractContent(XMLTag, Buffer, Index, Preformatted);
         XMLTag = XMLTag->Next;
      }
   }
}

/*****************************************************************************
** Function: InsertLineTag()
**
** Tag attributes are arranged as a byte array, in the following format:
**
**    LONG Size; { LONG Tag; LONG Size; BYTE Data[]; }
*/

static ERROR InsertLineTag(struct Text *Self, LONG TagID, LONG Line, STRING Value)
{
   BYTE *data;
   LONG i, len;

   if (Value) {
      for (len=0; Value[len]; len++);
      len++; /* NULL byte */
      if (len & 1) len++; /* Byte alignment */
      len += 8;
   }
   else len = 8;

   /*** Allocate a new tag array at the correct length ***/

   if (!Self->Array[Line].Tags) {
      if (AllocMemory(4 + len, MEM_DATA, &Self->Array[Line].Tags, NULL) != ERR_Okay) {
         return(ERR_AllocMemory);
      }
   }
   else if (ReallocMemory(Self->Array[Line].Tags, ((LONG *)Self->Array[Line].Tags)[0] + 4 + len, &Self->Array[Line].Tags, NULL) != ERR_Okay) {
      return(ERR_ReallocMemory);
   }

   /*** Rebuild our tag array data ***/

   i = sizeof(LONG) + ((LONG *)Self->Array[Line].Tags)[0];
   data = Self->Array[Line].Tags + i;
   ((LONG *)Self->Array[Line].Tags)[0] += len; /* Increase the total Size */
   ((LONG *)data)[0] = TagID;   /* Tag */
   ((LONG *)data)[1] = len;     /* Size */
   if (Value) {
      data += 8;
      for (i=0; Value[i]; i++) data[i] = Value[i];
      data[i] = 0;
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: CaclulateHScroll()
*/

static ERROR CalculateHScroll(struct Text *Self)
{
   struct mtUpdateScroll scroll;
   LONG width;

   if (!Self->HScroll) return(ERR_Okay);
   if (Self->NoUpdate) return(ERR_Okay);

   /*** If wordwrap is enabled then the horizontal scrollbar is pointless ***/

   if (Self->Flags & TXF_WORDWRAP) return(ERR_Okay);

   GET_TextWidth(Self, &width);
   scroll.ViewSize = -1;
   scroll.PageSize = width + Self->LeftMargin + Self->RightMargin;
   scroll.Position = -Self->XPosition;
   scroll.Unit     = Self->Font->MaxHeight;
   return(ActionMsg(MT_UpdateScroll, Self->HScroll, &scroll));
}

/*****************************************************************************
** Internal: CaclulateVScroll()
*/

static ERROR CalculateVScroll(struct Text *Self)
{
   struct mtUpdateScroll scroll;
   LONG lines, row, pagewidth;

   if (!Self->VScroll) return(ERR_Okay);
   if (Self->NoUpdate) return(ERR_Okay);

   if ((Self->Flags & TXF_WORDWRAP) AND (Self->AmtLines > 0) AND (Self->DrawableWidth > 0)) {
      pagewidth = Self->AreaWidth;

      lines = 0;
      pagewidth = pagewidth - Self->LeftMargin - Self->RightMargin;
      for (row=0; row < Self->AmtLines; row++) {
         if (Self->Array[row].PixelLength > pagewidth) {
            lines += ((Self->Array[row].PixelLength + pagewidth - 1) / pagewidth);
         }
         else lines++;
      }
   }
   else lines = Self->AmtLines;

   scroll.ViewSize = -1;
   scroll.PageSize = (lines * Self->Font->LineSpacing) + Self->AreaY + Self->TopMargin + Self->BottomMargin;
   scroll.Position = -Self->YPosition;
   scroll.Unit     = Self->Font->LineSpacing;
   return(ActionMsg(MT_UpdateScroll, Self->VScroll, &scroll));
}

/*****************************************************************************
** Internal: CalculateWidth()
*/

static LONG CalculateWidth(struct Text *Self, STRING String, LONG Length)
{
   struct mtCalcStringWidth calc;
   WORD i;

   if ((!Self) OR (!String)) return(NULL);

   if (Self->Flags & TXF_SECRET) {
      if (!Length) for (Length=0; String[Length]; Length++);

      BYTE buffer[Length+1];

      for (i=0; i < Length; i++) buffer[i] = '*';
      buffer[i] = 0;

      calc.String = buffer;
      calc.Length = Length;
      if (Action(MT_CalcStringWidth, Self->Font, &calc) IS ERR_Okay) {
         return(calc.Width);
      }
      else return(NULL);
   }
   else {
      if (!Length) Length = -1;
      calc.String = String;
      calc.Length = Length;
      if (Action(MT_CalcStringWidth, Self->Font, &calc) IS ERR_Okay) {
         return(calc.Width);
      }
      else return(NULL);
   }
}

/*****************************************************************************
** Function: CheckTag()
*/

static ERROR CheckTag(UBYTE *Tag, LONG TagID, STRING *Attribute)
{
   UBYTE *end;

   if (Attribute) *Attribute = NULL;

   if (Tag) {
      end = Tag + ((LONG *)Tag)[0] + sizeof(LONG);
      Tag += sizeof(LONG);
      while (Tag < end) {
         if (TagID IS ((LONG *)Tag)[0]) {
            if (Attribute) {
               if ((((LONG *)Tag)[1] > 8)) *Attribute = Tag + 8;
            }
            return(ERR_Okay);
         }
         if (!((LONG *)Tag)[1]) break;
         else Tag += ((LONG *)Tag)[1];
      }
      return(ERR_Search);
   }
   else return(ERR_Search);
}

/*****************************************************************************
** Internal: ColumnCoordinate()
**
** This function returns the exact horizontal coordinate for a specific column.
** The coordinate is absolute and relative to the text object's drawable
** container.
*/

static LONG ColumnCoordinate(struct Text *Self, LONG Row, LONG Column)
{
   struct mtCalcStringWidth calc;
   LONG alignx;
   WORD i;
   STRING str;
   BYTE buffer[Self->Array[Row].Length+1];

   alignx = 0;

   if (Self->Flags & TXF_SECRET) {
      str = buffer;
      for (i=0; i < Self->Array[Row].Length; i++) buffer[i] = '*';
      buffer[i] = 0;
   }
   else str = Self->Array[Row].String;

   if (Self->Font->AlignFlags & ALIGN_HORIZONTAL) {
      calc.String = str;
      calc.Length = -1;
      if (Action(MT_CalcStringWidth, Self->Font, &calc) IS ERR_Okay) {
         alignx = (Self->AreaWidth - calc.Width) / 2;
      }
   }

   if (Column <= 0) return(Self->AreaX + alignx + Self->LeftMargin + Self->XPosition);
   if (Column >= Self->Array[Row].Length) return(Self->AreaX + alignx + Self->LeftMargin + Self->XPosition + Self->Array[Row].PixelLength);

   if (Row >= Self->AmtLines) Row = Self->AmtLines - 1;

   if (str) {
      calc.String = str;
      calc.Length = Column;
      if (Action(MT_CalcStringWidth, Self->Font, &calc) IS ERR_Okay) {
         return(Self->AreaX + alignx + Self->LeftMargin + Self->XPosition + calc.Width);
      }
   }

   return(Self->AreaX + alignx + Self->LeftMargin + Self->XPosition);
}

/*****************************************************************************
** Internal: DeleteSelectedArea()
*/

static void DeleteSelectedArea(struct Text *Self)
{
   STRING str;
   LONG row, column, endrow, endcolumn, i, j;

   Self->Flags &= ~TXF_AREASELECTED;

   GetSelectedArea(Self, &row, &column, &endrow, &endcolumn);
   column = UTF8CharacterOffset(Self->Array[row].String, column);
   endcolumn = UTF8CharacterOffset(Self->Array[endrow].String, endcolumn);

   if (row IS endrow) {
      str = Self->Array[row].String;
      for (i=column; endcolumn < Self->Array[row].Length; i++) {
         str[i] = str[endcolumn++];
      }
      Self->Array[row].Length = i;
      Self->Array[row].String[i] = 0;
      Self->Array[row].PixelLength = CalculateWidth(Self, str, i);
      MoveCursor(Self, row, column);
      RedrawLine(Self, row);
      CalculateHScroll(Self);
   }
   else {
      i = column + (Self->Array[endrow].Length - endcolumn) + 1;
      if (AllocMemory(i, MEM_STRING|MEM_NOCLEAR, &str, NULL) IS ERR_Okay) {

         if (Self->Array[row].String) for (i=0; i < column; i++) str[i] = Self->Array[row].String[i];
         else i = 0;
         if (Self->Array[endrow].String) for (j=endcolumn; j < Self->Array[endrow].Length; j++) str[i++] = Self->Array[endrow].String[j];
         str[i] = 0;

         if (Self->Array[row].String) FreeMemory(Self->Array[row].String);
         Self->Array[row].String = str;
         Self->Array[row].Length = i;
         Self->Array[row].PixelLength = CalculateWidth(Self, str, i);
         MoveCursor(Self, row, column);

         /*** Delete following strings ***/

         row++;
         endrow++;
         for (i=row; i < endrow; i++) {
            if (Self->Array[i].String) FreeMemory(Self->Array[i].String);
         }

         for (i=row; endrow < Self->AmtLines; i++) {
            Self->Array[i].String      = Self->Array[endrow].String;
            Self->Array[i].PixelLength = Self->Array[endrow].PixelLength;
            Self->Array[i].Length      = Self->Array[endrow].Length;
            endrow++;
         }

         Self->AmtLines = i;

         DrawLines(Self, row-1, 30000);
         CalculateHScroll(Self);
         CalculateVScroll(Self);
      }
   }
}

/*****************************************************************************
** Internal: DrawLines()
*/

static void DrawLines(struct Text *Self, LONG Row, LONG Total)
{
   struct acDraw draw;
   LONG lines;

   if (Self->NoUpdate) return;

   if (Total < 1) return;

   if ((Self->Flags & TXF_WORDWRAP) AND (Row < Self->AmtLines)) {
      if (Row IS Self->AmtLines-1) {
         /*** Draw only the last word-wrapped line for speed ***/
         if (Self->Array[Row].PixelLength > 0) {
            lines = (Self->Array[Row].PixelLength + Self->AreaWidth - 1) / Self->AreaWidth;
            if (lines < 1) lines = 1;
         }
         else lines = Total;
      }
      else lines = 1000; /* Draw everything past the row to be redrawn */
   }
   else lines = Total;

   draw.XCoord = 0;
   draw.YCoord = RowCoordinate(Self, Row);
   draw.Width  = 30000;
   draw.Height = lines * Self->Font->LineSpacing;
   ActionMsg(AC_Draw, Self->DrawableID, &draw);
}

/*****************************************************************************
** Internal: RedrawLine()
*/

static void RedrawLine(struct Text *Self, LONG Line)
{
   struct acDraw draw;
   LONG lines;

   if (Self->NoUpdate) return;
   if (Line < 0) return;

   if ((Self->Flags & TXF_WORDWRAP) AND (Line < Self->AmtLines)) {
      if (Self->Array[Line].PixelLength > 0) {
         lines = (Self->Array[Line].PixelLength + Self->AreaWidth - 1) / Self->AreaWidth;
         if (lines < 1) lines = 1;
      }
      else lines = 1;
   }
   else lines = 1;

   draw.XCoord = 0;
   draw.YCoord = RowCoordinate(Self, Line);
   draw.Height = Self->Font->LineSpacing * lines;
   draw.Width  = 30000;
   ActionMsg(AC_Draw, Self->DrawableID, &draw);
}

/*****************************************************************************
** Internal: GetSelectedArea()
*/

static void GetSelectedArea(struct Text *Self, LONG *Row, LONG *Column, LONG *EndRow, LONG *EndColumn)
{
   if (Self->SelectRow < Self->CursorRow) {
      *Row       = Self->SelectRow;
      *EndRow    = Self->CursorRow;
      *Column    = Self->SelectColumn;
      *EndColumn = Self->CursorColumn;
   }
   else if (Self->SelectRow IS Self->CursorRow) {
      *Row       = Self->SelectRow;
      *EndRow    = Self->CursorRow;
      if (Self->SelectColumn < Self->CursorColumn) {
         *Column    = Self->SelectColumn;
         *EndColumn = Self->CursorColumn;
      }
      else {
         *Column    = Self->CursorColumn;
         *EndColumn = Self->SelectColumn;
      }
   }
   else {
      *Row       = Self->CursorRow;
      *EndRow    = Self->SelectRow;
      *Column    = Self->CursorColumn;
      *EndColumn = Self->SelectColumn;
   }
}

/*****************************************************************************
** Internal: MoveCursor()
*/

static void MoveCursor(struct Text *Self, LONG Row, LONG Column)
{
   /* If an area is currently selected, turn off the selection
   ** and redraw the graphics area.
   */

   if (Self->Flags & TXF_AREASELECTED) {
      Self->Flags &= ~TXF_AREASELECTED;
      Redraw(Self);
   }

   if (Row < 0) Row = 0;
   else if (Row >= Self->AmtLines) Row = Self->AmtLines - 1;

   if (Column < 0) Column = 0;

   /*** Remove the current cursor first ***/

   RemoveCursor(Self);

   /*** Make sure that the cursor is going to be viewable at its new position ***/

   Self->CursorRow = Row;
   Self->CursorColumn = Column;
   ViewCursor(Self);

   /*** Redraw the cursor at its new position ***/

   RedrawCursor(Self, TRUE);
}

/*****************************************************************************
** Internal: RedrawCursor()
** Short:    Redraws the cursor area.
*/

static void RedrawCursor(struct Text *Self, LONG Visible)
{
   struct acDraw draw;

   if (Self->NoUpdate) return;

   if (!(Self->Flags & TXF_EDIT)) return;

   #ifdef DBG_TEXT
      DPrintF("~RedrawCursor()","Visible: %d", Visible);
   #endif

   if (Visible IS FALSE) Self->NoCursor++;

   draw.XCoord = ColumnCoordinate(Self, Self->CursorRow, Self->CursorColumn);
   draw.YCoord = RowCoordinate(Self, Self->CursorRow);
   draw.Height = Self->Font->LineSpacing;
   draw.Width  = Self->CursorWidth + 1;

   #ifdef DBG_TEXT
      DPrintF("RedrawCursor:","%dx%d, %dx%d", (LONG)draw.XCoord, (LONG)draw.YCoord, (LONG)draw.Width, (LONG)draw.Height);
   #endif

   ActionMsg(AC_Draw, Self->DrawableID, &draw);

   if (Visible IS FALSE) Self->NoCursor--;

   #ifdef DBG_TEXT
      StepBack();
   #endif
}

/*****************************************************************************
** Internal: InsertCharacter()
*/

static void InsertCharacter(struct Text *Self, LONG Unicode, LONG Column)
{
   STRING str;
   LONG i, j, k, charlen, offset, unicodelen;
   UBYTE buffer[6];

   if ((!Self) OR (!Unicode)) return;
   
   /*** If FORCECAPS is used, convert lower case letters to upper case ***/

   if (Self->Flags & TXF_FORCECAPS) {
      if ((Unicode >= 'a') AND (Unicode <= 'z')) Unicode = Unicode - 'a' + 'A';
   }

   /*** Convert the character into a UTF-8 sequence ***/

   charlen = UTF8WriteValue(Unicode, buffer, 6);

   if ((!Self->Array[Self->CursorRow].String) OR (Self->Array[Self->CursorRow].Length < 1)) {
      if (Self->CharLimit < 1) return;

      if (Self->Array[Self->CursorRow].String) FreeMemory(Self->Array[Self->CursorRow].String);

      if (AllocMemory(charlen+1, MEM_STRING|MEM_NOCLEAR, &Self->Array[Self->CursorRow].String, NULL) IS ERR_Okay) {
         for (i=0; i < charlen; i++) Self->Array[Self->CursorRow].String[i] = buffer[i];
         Self->Array[Self->CursorRow].String[i]   = 0;
         Self->Array[Self->CursorRow].Length = charlen;
         Self->Array[Self->CursorRow].PixelLength = CalculateWidth(Self, Self->Array[Self->CursorRow].String, 1);

         Self->CursorColumn = 1;

         if (Self->AmtLines <= Self->CursorRow) Self->AmtLines = Self->CursorRow + 1;

         RedrawLine(Self, Self->CursorRow);
         CalculateHScroll(Self);
      }
   }
   else {
      offset = UTF8CharacterOffset(Self->Array[Self->CursorRow].String, Column);

      if (offset > Self->Array[Self->CursorRow].Length) return;

      if (AllocMemory(Self->Array[Self->CursorRow].Length + charlen + 1, MEM_STRING|MEM_NOCLEAR, (APTR *)&str, NULL) IS ERR_Okay) {
         j = 0;
         i = 0;
         while (i < offset) str[j++] = Self->Array[Self->CursorRow].String[i++]; /* Copy existing characters */
         for (k=0; k < charlen; k++) str[j++] = buffer[k]; /* Insert new character */

         /* If overwrite mode is set, skip over the character bytes at the
         ** current cursor position.
         */

         if ((Self->Flags & TXF_OVERWRITE) AND (i < Self->Array[Self->CursorRow].Length)) {
            charlen--;
            for (++i; (Self->Array[Self->CursorRow].String[i] & 0xc0) IS 0x80; i++) charlen--;
         }

         while (i < Self->Array[Self->CursorRow].Length) str[j++] = Self->Array[Self->CursorRow].String[i++]; /* Copy remaining characters */
         str[j] = 0;

         FreeMemory(Self->Array[Self->CursorRow].String);
         Self->Array[Self->CursorRow].String = str;
         Self->Array[Self->CursorRow].Length += charlen;
         Self->CursorColumn++;

         /*** Get the UTF-8 length of this string so that we can enforce character limits ***/

         for (unicodelen=0, i=0; i < Self->Array[Self->CursorRow].Length; unicodelen++) {
            for (++i; (str[i] & 0xc0) IS 0x80; i++);
         }

         if (unicodelen > Self->CharLimit) {
            /*** Delete the character that is at the end of this line to keep it within limits ***/
            Self->Array[Self->CursorRow].Length -= UTF8PrevLength(str, Self->Array[Self->CursorRow].Length);
            str[Self->Array[Self->CursorRow].Length] = 0;
         }

         /*** Ensure that the cursor column does not exceed the length of the line ***/

         if (Self->CursorColumn > unicodelen) Self->CursorColumn = unicodelen;

         Self->Array[Self->CursorRow].PixelLength = CalculateWidth(Self, str, Self->Array[Self->CursorRow].Length);

         RedrawLine(Self, Self->CursorRow);

         if (ViewCursor(Self) IS FALSE) CalculateHScroll(Self);
      }
   }
}

/*****************************************************************************
** Internal: EditKeypress()
*/

static ERROR EditKeypress(struct Text *Self, LONG Flags, LONG Value)
{
   struct acClipboard clip;
   struct dcKeyEntry keys[2];
   UBYTE string[6];
   OBJECTPTR drawable, keyboard;
   STRING str;
   LONG i, j, row, len, offset;

   Self->CursorFlash = 0; /* Reset the flashing cursor to make it visible */

   if ((!(Self->Flags & TXF_NOSYSKEYS)) AND (Flags & KQ_CTRL)) {
      switch(Value) {
         case K_C: /*** Copy ***/
            clip.ClipboardID = glClipboardID;
            clip.Mode        = CLIP_COPY;
            Action(AC_Clipboard, Self, &clip);
            return(ERR_Okay);

         case K_X: /*** Cut ***/
            if (!(Self->Flags & TXF_EDIT)) return(ERR_Okay);
            clip.ClipboardID = glClipboardID;
            clip.Mode        = CLIP_CUT;
            Action(AC_Clipboard, Self, &clip);
            return(ERR_Okay);

         case K_V: /*** Paste ***/
            if (!(Self->Flags & TXF_EDIT)) return(ERR_Okay);
            clip.ClipboardID = glClipboardID;
            clip.Mode        = CLIP_PASTE;
            Action(AC_Clipboard, Self, &clip);
            return(ERR_Okay);

         case K_K: /*** Delete line ***/
            if (!(Self->Flags & TXF_EDIT)) return(ERR_Okay);
            ActionTags(MT_DeleteTextLine, Self, Self->CursorRow);
            return(ERR_Okay);

         case K_Z: /*** Undo ***/
            if (!(Self->Flags & TXF_EDIT)) return(ERR_Okay);
            return(ERR_Okay);

         case K_Y: /*** Redo ***/
            if (!(Self->Flags & TXF_EDIT)) return(ERR_Okay);
            return(ERR_Okay);
      }
   }

   if (!(Flags & KQ_NOTPRINTABLE)) { // AND (!(Flags & KQ_INSTRUCTIONKEYS))

      /*** Printable character handling ***/

      if (!(Self->Flags & TXF_EDIT)) return(ERR_Okay);

      keys[0].Flags = Flags;
      keys[0].Value = Value;
      keys[1].Flags = NULL;
      keys[1].Value = NULL;

      if (AccessObject(Self->InputID, 5000, &keyboard) IS ERR_Okay) {
         if (ActionTags(MT_TranslateKeys, keyboard, keys, string, 6) IS ERR_Okay) {
            ReleaseObject(keyboard);
            if (string[0]) {
               if (Self->Flags & TXF_AREASELECTED) DeleteSelectedArea(Self);
               InsertCharacter(Self, UTF8ReadValue(string, NULL), Self->CursorColumn);
            }
         }
         else ReleaseObject(keyboard);
      }

      return(ERR_Okay);
   }

   if (!(Self->Flags & TXF_EDIT)) return(ERR_Okay);

   switch(Value) {
   case K_BACKSPACE:
      if (Self->Flags & TXF_AREASELECTED) DeleteSelectedArea(Self);
      else if (Self->CursorColumn > 0) {
         if (Self->CursorColumn >= Self->Array[Self->CursorRow].Length) {
            Self->CursorColumn = Self->Array[Self->CursorRow].Length - 1;
         }
         else Self->CursorColumn--;

         if ((str = Self->Array[Self->CursorRow].String)) {
            i = UTF8CharacterOffset(str, Self->CursorColumn);
            for (len=1; (str[i+len] & 0xc0) IS 0x80; len++);
            for (; str[i+len]; i++) str[i] = str[i+len];
            str[i] = 0;
            Self->Array[Self->CursorRow].Length--;
            Self->Array[Self->CursorRow].PixelLength = CalculateWidth(Self, str, Self->Array[Self->CursorRow].Length);
            RedrawLine(Self, Self->CursorRow);
            CalculateHScroll(Self);
         }
      }
      else if (Self->CursorRow > 0) {
         /*** This routine is used if the current line will be shifted up into the line above it ***/
         len = Self->Array[Self->CursorRow-1].Length + Self->Array[Self->CursorRow].Length;
         if (len > 0) {
            if (AllocMemory(len+1, MEM_STRING|MEM_NOCLEAR, &str, NULL) IS ERR_Okay) {
               for (i=0; i < Self->Array[Self->CursorRow-1].Length; i++) str[i] = Self->Array[Self->CursorRow-1].String[i];
               for (j=0; j < Self->Array[Self->CursorRow].Length; j++) str[i++] = Self->Array[Self->CursorRow].String[j];
               str[i] = 0;

               Self->CursorRow--;
               Self->CursorColumn = UTF8Length(Self->Array[Self->CursorRow].String);
               if (Self->Array[Self->CursorRow].String) FreeMemory(Self->Array[Self->CursorRow].String);
               Self->Array[Self->CursorRow].String = str;
               Self->Array[Self->CursorRow].Length = len;
               Self->Array[Self->CursorRow].PixelLength = CalculateWidth(Self, str, len);
               RedrawLine(Self, Self->CursorRow);
               ActionTags(MTX_DeleteLine, Self, Self->CursorRow+1);
               ViewCursor(Self);
            }
         }
         else {
            ActionTags(MTX_DeleteLine, Self, Self->CursorRow);
            MoveCursor(Self, Self->CursorRow-1, Self->Array[Self->CursorRow-1].Length);
         }
      }
      break;

   case K_CLEAR:
      if (Self->Flags & TXF_AREASELECTED) DeleteSelectedArea(Self);
      else {
         Self->CursorColumn = 0;
         ActionTags(MTX_DeleteLine, Self, Self->CursorRow);
      }
      break;

   case K_DELETE:
      if (Self->Flags & TXF_AREASELECTED) DeleteSelectedArea(Self);
      else if (Self->CursorColumn < Self->Array[Self->CursorRow].Length) {
         if ((str = Self->Array[Self->CursorRow].String)) {
            offset = UTF8CharacterOffset(Self->Array[Self->CursorRow].String, Self->CursorColumn);
            len = UTF8CharacterLength(Self->Array[Self->CursorRow].String+offset);
            for (i=offset; str[i+len]; i++) str[i] = str[i+len];
            str[i] = 0;
            Self->Array[Self->CursorRow].Length -= len;
            Self->Array[Self->CursorRow].PixelLength = CalculateWidth(Self, str, Self->Array[Self->CursorRow].Length);
            RedrawLine(Self, Self->CursorRow);
            CalculateHScroll(Self);
         }
      }
      else if (Self->CursorRow < Self->AmtLines - 1) {
         /*** This code is used if the next line is going to be pulled up into the current line ***/
         len = Self->Array[Self->CursorRow+1].Length + Self->Array[Self->CursorRow].Length;
         if (len > 0) {
            if (AllocMemory(len+1, MEM_STRING|MEM_NOCLEAR, &str, NULL) IS ERR_Okay) {
               for (i=0; i < Self->Array[Self->CursorRow].Length; i++) str[i] = Self->Array[Self->CursorRow].String[i];
               for (j=0; j < Self->Array[Self->CursorRow+1].Length; j++) str[i++] = Self->Array[Self->CursorRow+1].String[j];
               str[i] = 0;

               if (Self->Array[Self->CursorRow].String) FreeMemory(Self->Array[Self->CursorRow].String);
               Self->Array[Self->CursorRow].String = str;
               Self->Array[Self->CursorRow].Length = len;
               Self->Array[Self->CursorRow].PixelLength = CalculateWidth(Self, str, len);
               RedrawLine(Self, Self->CursorRow);
               ActionTags(MTX_DeleteLine, Self, Self->CursorRow+1);
            }
         }
         else ActionTags(MTX_DeleteLine, Self, Self->CursorRow);
      }
      break;

   case K_END:
      MoveCursor(Self, Self->CursorRow, Self->Array[Self->CursorRow].Length);
      break;

   case K_ENTER:
   case K_NPENTER:
      Self->HistoryPos = 0; /* Reset the history position */
      
      if (Self->Flags & TXF_ENTERTAB) {
         /* Match the enter-key with an emulated tab-key press (this is
         ** useful for things like input boxes).  Note that we grab the
         ** keyboard exclusively to send the tab at this exact moment.
         ** Otherwise the delay can cause the keypress to be sent to the
         ** wrong subscribers if the current window closes for example.
         */
         struct dcKeyEntry key;
         OBJECTPTR input;
         key.Flags = KQ_PRESSED;
         key.Value = K_TAB; 
         if (AccessObject(Self->InputID, 3000, &input) IS ERR_Okay) {
            acDataChannel(input, Self->Head.UniqueID, DATA_KEYPRESS, 1, &key, sizeof(key), 1);
            ReleaseObject(input);
         }
      }
      
      if ((Self->Flags & TXF_ENTERACTIVATE) OR (Self->SelectionObjectID)) {
         if (Self->Flags & TXF_ENTERACTIVATE) Action(AC_Activate, Self, NULL);

         if (Self->TabFocusID) {
            Action(AC_LostFocus, Self, NULL);
            ActionMsg(AC_Focus, Self->TabFocusID, NULL);
         }
         else if (Self->SelectionObjectID) {
            OBJECTPTR object;
            if (AccessObject(Self->SelectionObjectID, 5000, &object) IS ERR_Okay) {
               if (Self->SelectionField[0]) {
                  SetFieldVariable(object, Self->SelectionField, Self->Array[0].String);
               }
               else SetField(object, FID_String, FT_STRING, Self->Array[0].String);
               ReleaseObject(object);
            }
         }

         if (Self->Flags & TXF_AUTOCLEAR) {
            if (Self->Flags & TXF_HISTORY) AddHistory(Self, Self->Array[0].String);
            SetField(Self, FID_String, FT_STRING, "");
            Self->CursorColumn = 0;
         }
      }
      else {
         if ((Self->LineLimit) AND (Self->AmtLines >= Self->LineLimit)) break;

         if (Self->Flags & TXF_AREASELECTED) DeleteSelectedArea(Self);

         if (!Self->AmtLines) {
            Self->AmtLines = 1;
            Self->Array[0].String = NULL;
            Self->Array[0].Length = NULL;
            Self->Array[0].PixelLength = NULL;
         }

         row    = Self->CursorRow;
         offset = UTF8CharacterOffset(Self->Array[row].String, Self->CursorColumn);
         Self->CursorRow++;
         Self->CursorColumn = 0;
         AddLine(Self, Self->Array[row].String + offset, row+1, Self->Array[row].Length - offset, FALSE);

         if (!offset) ActionTags(MTX_ReplaceLine, Self, NULL, row, 0);
         else ActionTags(MTX_ReplaceLine, Self, Self->Array[row].String, row, offset);
         ViewCursor(Self);
      }
      break;

   case K_HOME:
      MoveCursor(Self, Self->CursorRow, 0);
      break;

   case K_INSERT:
      if (Self->Flags & TXF_OVERWRITE) Self->Flags &= ~TXF_OVERWRITE;
      else Self->Flags |= TXF_OVERWRITE;
      break;

   case K_LEFT:
      ValidateCursorPosition(Self, FALSE);
      if (Self->CursorColumn > 0) MoveCursor(Self, Self->CursorRow, Self->CursorColumn-1);
      else if (Self->CursorRow > 0) MoveCursor(Self, Self->CursorRow-1, UTF8Length(Self->Array[Self->CursorRow-1].String));
      break;

   case K_PAGEDOWN:
      if (Self->LineLimit IS 1) break;

      if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {
         if (GetField(drawable, FID_Height, FT_LONG, &i) IS ERR_Okay) {
            MoveCursor(Self, Self->CursorRow + (i/(Self->Font->LineSpacing)), Self->CursorColumn);
         }
         ReleaseObject(drawable);
      }
      break;

   case K_PAGEUP:
      if (Self->LineLimit IS 1) break;

      if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {
         if (GetField(drawable, FID_Height, FT_LONG, &i) IS ERR_Okay) {
            MoveCursor(Self, Self->CursorRow - (i/(Self->Font->LineSpacing)), Self->CursorColumn);
         }
         ReleaseObject(drawable);
      }
      break;

   case K_RIGHT:
      ValidateCursorPosition(Self, FALSE);
      if (Self->CursorColumn < UTF8Length(Self->Array[Self->CursorRow].String)) {
         MoveCursor(Self, Self->CursorRow, Self->CursorColumn+1);
      }
      else if (Self->CursorRow < Self->AmtLines - 1) {
         MoveCursor(Self, Self->CursorRow+1, 0);
      }
      break;

   case K_TAB:
      if (Self->TabFocusID) {
         Action(AC_LostFocus, Self, NULL);
         ActionMsg(AC_Focus, Self->TabFocusID, NULL);
      }
      else {
         if ((Self->Flags & TXF_TABKEY) OR (Flags & KQ_SHIFT)) {
            if (Self->Flags & TXF_AREASELECTED) DeleteSelectedArea(Self);
            InsertCharacter(Self, '\t', Self->CursorColumn);
         }
      }
      break;

   case K_DOWN:
   case K_UP:
      if (Self->Flags & TXF_HISTORY) {
         if (Value IS K_UP) {
            /*** Return if we area already at the maximum historical position ***/
            if (Self->HistoryPos >= Self->HistorySize) return(ERR_Okay);
            Self->HistoryPos++;
            for (i=0; i < Self->HistorySize; i++) {
               if (Self->History[i].Number IS Self->HistoryPos) {
                  SetField(Self, FID_String, FT_STRING, Self->History[i].Buffer);
                  for (Self->CursorColumn=0; Self->History[i].Buffer[Self->CursorColumn]; Self->CursorColumn++);
                  Redraw(Self);
                  return(ERR_Okay);
               }
            }
            /* If we couldn't find a string to match the higher history
            ** number, revert to what it was.
            */
            Self->HistoryPos--;
         }
         else {
            if (Self->HistoryPos <= 0) {
               SetField(Self, FID_String, FT_STRING, "");
               Self->CursorColumn = 0;
               return(ERR_Okay);
            }
            Self->HistoryPos--;
            for (i=0; i < Self->HistorySize; i++) {
               if (Self->History[i].Number IS Self->HistoryPos) {
                  SetField(Self, FID_String, FT_STRING, Self->History[i].Buffer);
                  for (Self->CursorColumn=0; Self->History[i].Buffer[Self->CursorColumn]; Self->CursorColumn++);
                  Redraw(Self);
                  return(ERR_Okay);
               }
            }
         }
      }
      else if (((Value IS K_UP) AND (Self->CursorRow > 0)) OR
               ((Value IS K_DOWN) AND (Self->CursorRow < Self->AmtLines-1))) {
         WORD endcolumn, col, colchar;

         /* Determine the current true position of the current cursor column,
         ** in UTF-8, with respect to tabs.  Then determine the cursor
         ** character that we are going to be at when we end up at the
         ** row above us.
         */

         if (((Self->CursorRow << 16) | Self->CursorColumn) IS Self->CursorSavePos) endcolumn = Self->CursorEndColumn;
         else {
            Self->CursorEndColumn = 0;
            endcolumn = Self->CursorColumn;
         }

         colchar = 0;
         col = 0;
         i = 0;
         while ((i < Self->Array[Self->CursorRow].Length) AND (colchar < endcolumn)) {
            if (Self->Array[Self->CursorRow].String[i] IS '\t') {
               col += ROUNDUP(col, Self->Font->TabSize);
            }
            else col++;

            colchar++;

            for (++i; (Self->Array[Self->CursorRow].String[i] & 0xc0) IS 0x80; i++);
         }

         /* If an area is currently selected, turn off the selection
         ** and redraw the graphics area.
         */

         if (Self->Flags & TXF_AREASELECTED) {
            Self->Flags &= ~TXF_AREASELECTED;
            Redraw(Self);
         }

         /*** Remove the current cursor first ***/

         RemoveCursor(Self);

         if (Value IS K_UP) Self->CursorRow--;
         else Self->CursorRow++;

         for (Self->CursorColumn=0, i=0; (col > 0) AND (i < Self->Array[Self->CursorRow].Length);) {
            if (Self->Array[Self->CursorRow].String[i] IS '\t') {
               col -= ROUNDUP(Self->CursorColumn, Self->Font->TabSize);
               Self->CursorColumn++;
            }
            else {
               col--;
               Self->CursorColumn++;
            }

            for (++i; (Self->Array[Self->CursorRow].String[i] & 0xc0) IS 0x80; i++);
         }

         if (Self->CursorColumn > Self->CursorEndColumn) Self->CursorEndColumn = Self->CursorColumn;

         Self->CursorSavePos = ((LONG)Self->CursorRow << 16) | Self->CursorColumn;

         /*** Make sure that the cursor is going to be viewable at its new position ***/

         ViewCursor(Self);
         RedrawCursor(Self, TRUE);
      }
      break;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: Redraw()
*/

static void Redraw(struct Text *Self)
{
   struct acDraw draw;

   if (Self->NoUpdate) return;

   draw.XCoord = Self->AreaX;
   draw.YCoord = Self->AreaY;
   draw.Width  = Self->AreaWidth;
   draw.Height = Self->AreaHeight;
   ActionMsg(AC_Draw, Self->DrawableID, &draw);
}

/*****************************************************************************
** Internal: ReplaceLine()
*/

static ERROR ReplaceLine(struct Text *Self, STRING String, LONG Line, LONG ByteLength)
{
   STRING str;
   LONG len, i, unicodelen;

   if ((Line < 0) OR (Line >= Self->AmtLines)) return(ERR_Args);

   /*** Calculate the length of the text if necessary ***/

   len = 0;
   if ((String) AND (*String)) {
      if (ByteLength >= 0) len = ByteLength;
      else for (len=0; (String[len]) AND (String[len] != '\n') AND (String[len] != '\r'); len++);
   }

   /*** Stop the string from exceeding the acceptable character limit ***/

   if (len >= Self->CharLimit) {
      for (unicodelen=0, i=0; (i < len) AND (unicodelen < Self->CharLimit); unicodelen++) {
         for (++i; (String[i] & 0xc0) IS 0x80; i++);
      }
      len = i;
   }

   if (len < 1) {
      /*** If the length is zero, clear the line ***/
      if (Self->Array[Line].String) FreeMemory(Self->Array[Line].String);
      if (Self->Array[Line].Tags)   FreeMemory(Self->Array[Line].Tags);
      Self->Array[Line].String      = NULL;
      Self->Array[Line].Tags        = NULL;
      Self->Array[Line].Length      = NULL;
      Self->Array[Line].PixelLength = NULL;
   }
   else if (len <= Self->Array[Line].Length) {
      /* If the new string is smaller than the available space,
      ** copy the new string straight over the old one.
      */

      for (i=0; (i < len) AND (String[i]); i++) {
         Self->Array[Line].String[i] = String[i];
      }
      Self->Array[Line].String[i]   = 0;
      Self->Array[Line].Tags        = NULL;
      Self->Array[Line].Length      = i;
      Self->Array[Line].PixelLength = CalculateWidth(Self, String, i);
   }
   else if (AllocMemory(len+1, MEM_STRING|MEM_NOCLEAR, (APTR *)&str, NULL) IS ERR_Okay) {
      if (Self->Array[Line].String) FreeMemory(Self->Array[Line].String);
      if (Self->Array[Line].Tags)   FreeMemory(Self->Array[Line].Tags);
      for (i=0; (i < len) AND (String[i]); i++) str[i] = String[i];
      str[i] = 0;
      Self->Array[Line].String = str;
      Self->Array[Line].Tags   = NULL;
      Self->Array[Line].Length = i;
      Self->Array[Line].PixelLength = CalculateWidth(Self, String, i);
   }
   else return(ERR_AllocMemory);

   if (!Self->NoUpdate) {
      CalculateHScroll(Self);
      RedrawLine(Self, Line);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: RowCoordinate()
*/

static LONG RowCoordinate(struct Text *Self, LONG Row)
{
   LONG y, height, line, pagewidth, i;

   y = Self->AreaY;

   if (Self->Flags & TXF_WORDWRAP) {
      line = 0;
      pagewidth = Self->AreaWidth - Self->LeftMargin - Self->RightMargin;
      for (i=0; i < Row; i++) {
         if (Self->Array[i].PixelLength > pagewidth) {
            line += ((Self->Array[i].PixelLength + pagewidth - 1) / pagewidth);
         }
         else line++;
      }
   }
   else line = Row;

   if (Self->Font->AlignFlags & ALIGN_VERTICAL) {
      GET_TextHeight(Self, &height);

      y += (Self->AreaHeight - height) / 2;
   }

   y += Self->TopMargin + (line * Self->Font->LineSpacing) + Self->YPosition;

   return(y);
}

/*****************************************************************************
** Internal: ValidateCursorPosition()
**
** If the cursor is out of the current line's boundaries, this function will
** move it to a safe position.
*/

static void ValidateCursorPosition(struct Text *Self, LONG Redraw)
{
   LONG column;

   if ((!Self->Array[Self->CursorRow].String) OR (Self->Array[Self->CursorRow].Length < 1)) {
      column = 0;
   }
   else {
      column = UTF8Length(Self->Array[Self->CursorRow].String);
      if (Self->CursorColumn < column) column = Self->CursorColumn;
   }

   if (column != Self->CursorColumn) {
      if (Redraw IS TRUE) MoveCursor(Self, Self->CursorRow, column);
      else Self->CursorColumn = column;
   }
}

/*****************************************************************************
** Internal: ViewCursor()
** Short:    Returns TRUE if the cursor was out of sight and needed to be scrolled into view.
**
** Note: If the cursor is not giving enough space at the bottom or right areas
** of the text view, simply increase the bottom and right margin values so that
** the cursor appears before it reaches the edge.
*/

static LONG ViewCursor(struct Text *Self)
{
   LONG ycoord, height, xcoord, width, scrolly, scrollx, xpos, ypos;
   BYTE scroll;

   //DPrintF("ViewCursor()","[%d] Row: %d, Column: %d", Self->Head.UniqueID, Self->CursorRow, Self->CursorColumn);

   if ((Self->DrawableHeight < 1) OR (Self->DrawableWidth < 1)) return(FALSE);

   scroll = FALSE;
   xpos = Self->XPosition;
   ypos = Self->YPosition;

   /*** Vertical positioning for the cursor ***/

   ycoord  = RowCoordinate(Self, Self->CursorRow);
   scrolly = 0;

   if (ycoord < Self->AreaY) {
      if (!Self->CursorRow) scrolly = -ypos;
      else scrolly = -ycoord + Self->AreaY;
   }
   else {
      height = Self->DrawableHeight;
      if (height > Self->BottomMargin) height -= Self->BottomMargin; /* This compensates for any obscuring scrollbar and keeps the cursor inside the view */
      if ((ycoord + Self->Font->LineSpacing) > height) {
         scrolly = -((ycoord + Self->Font->LineSpacing) - height);
      }
   }

   /*** Horizontal positioning for the cursor ***/

   xcoord  = ColumnCoordinate(Self, Self->CursorRow, Self->CursorColumn);
   scrollx = 0;

   if (xcoord < Self->AreaX) {
      if (!Self->CursorColumn) scrollx = -xpos; /* Scroll to position zero */
      else scrollx = -xcoord + Self->AreaX;
   }
   else {
      width = Self->DrawableWidth;
      if (width > Self->RightMargin) width -= Self->RightMargin; /* This compensates for any obscuring scrollbar and keeps the cursor inside the view */
      if (xcoord > width) scrollx = -(xcoord - width);
   }

   /*** Do the scroll action ***/

   if ((scrollx) OR (scrolly)) {
      ActionTags(AC_ScrollToPoint, Self, (DOUBLE)((-xpos) - scrollx), (DOUBLE)((-ypos) - scrolly), 0.0, STF_X|STF_Y);

      if (!Self->NoUpdate) {
         if (scrollx) CalculateHScroll(Self);
         if (scrolly) CalculateVScroll(Self);
      }
      scroll = TRUE;
   }

   return(scroll);
}

/*****************************************************************************
** Internal: ViewSelection()
** Short:    Returns TRUE if the cursor was out of sight and needed to be scrolled into view.
**
** Similar to ViewCursor() but includes the selected area.  Intended for use
** by text highlighting functions, e.g. for finding text.
*/

static LONG ViewSelection(struct Text *Self)
{
   LONG ycoord, height, xcoord, width, scrolly, scrollx, selectx, selecty, xpos, ypos;
   BYTE scroll;

   if ((Self->DrawableHeight < 1) OR (Self->DrawableWidth < 1)) return(FALSE);

   scroll = FALSE;
   xpos = Self->XPosition;
   ypos = Self->YPosition;

   /*** Vertical positioning for the start of the select area ***/

   ycoord  = RowCoordinate(Self, Self->SelectRow);
   selecty = 0;

   if (ycoord < Self->AreaY) {
      if (!Self->SelectRow) selecty = -ypos;
      else selecty = -ycoord + Self->AreaY;
   }
   else {
      height = Self->DrawableHeight;
      if (height > Self->BottomMargin) height -= Self->BottomMargin; /* This compensates for any obscuring scrollbar and keeps the cursor inside the view */
      if ((ycoord + Self->Font->LineSpacing) > height) {
         selecty = -((ycoord + Self->Font->LineSpacing) - height);
      }
   }

   /*** Horizontal positioning for the start of the select area ***/

   xcoord  = ColumnCoordinate(Self, Self->SelectRow, Self->SelectColumn);
   selectx = 0;

   if (xcoord < Self->AreaX) {
      if (!Self->SelectColumn) selectx = -xpos; /* Scroll to position zero */
      else selectx = -xcoord + Self->AreaX;
   }
   else {
      width = Self->DrawableWidth;
      if (width > Self->RightMargin) width -= Self->RightMargin; /* This compensates for any obscuring scrollbar and keeps the cursor inside the view */
      if (xcoord > width) selectx = -(xcoord - width);
   }

   /*** Vertical positioning for the cursor ***/

   ycoord  = RowCoordinate(Self, Self->CursorRow);
   scrolly = 0;

   if (ycoord < Self->AreaY) {
      if (!Self->CursorRow) scrolly = -ypos;
      else scrolly = -ycoord + Self->AreaY;
   }
   else {
      height = Self->DrawableHeight;
      if (height > Self->BottomMargin) height -= Self->BottomMargin; /* This compensates for any obscuring scrollbar and keeps the cursor inside the view */
      if ((ycoord + Self->Font->LineSpacing) > height) {
         scrolly = -((ycoord + Self->Font->LineSpacing) - height);
      }
   }

   /*** Horizontal positioning for the cursor ***/

   xcoord  = ColumnCoordinate(Self, Self->CursorRow, Self->CursorColumn);
   scrollx = 0;

   if (xcoord < Self->AreaX) {
      if (!Self->CursorColumn) scrollx = -xpos; /* Scroll to position zero */
      else scrollx = -xcoord + Self->AreaX;
   }
   else {
      width = Self->DrawableWidth;
      if (width > Self->RightMargin) width -= Self->RightMargin; /* This compensates for any obscuring scrollbar and keeps the cursor inside the view */
      if (xcoord > width) scrollx = -(xcoord - width);
   }

   /*** Do the scroll action ***/

   if ((selectx) AND (!scrollx)) {
      scrollx = selectx;
   }

   if ((selecty) AND (!scrolly)) {
      scrolly = selecty;
   }

   if ((scrollx) OR (scrolly)) {
      ActionTags(AC_ScrollToPoint, Self, (DOUBLE)((-xpos) - scrollx), (DOUBLE)((-ypos) - scrolly), 0.0, STF_X|STF_Y);

      if (!Self->NoUpdate) {
         if (scrollx) CalculateHScroll(Self);
         if (scrolly) CalculateVScroll(Self);
      }
      scroll = TRUE;
   }

   return(scroll);
}

