#ifndef GUI_RENDER_H
#define GUI_RENDER_H 1

/*
**  render.h
**
**  (C) Copyright 2000-2004 Rocklyte Systems.
**      All Rights Reserved.
*/

#ifndef MAIN_H
#include <pandora/main.h>
#endif

/*****************************************************************************
** Render class.
*/

struct RenderCallback {
   LONG Data;
   APTR Routine; //void (*Routine)(OBJECTPTR, struct Render *, struct Bitmap *);
};

struct Render {
   OBJECT_HEADER            /* Standard header */
   OBJECTID DragID;         /* Drag the object that this field points to */
   OBJECTID BufferID;       /* Buffer bitmap (backing store) */
   OBJECTID ParentID;       /* Graphical container of the Render object, if any */
   OBJECTID RouteFocusID;   /* Allows you to re-route the focus trail in a different tree direction */
   LONG     TopMargin;      /* Top movement limit */
   LONG     BottomMargin;   /* Bottom movement limit */
   LONG     LeftMargin;     /* Left movement limit */
   LONG     RightMargin;    /* Right movement limit*/
   LONG     MinWidth;       /* Minimum width setting */
   LONG     MinHeight;      /* Minimum height setting */
   LONG     MaxWidth;       /* Maximum width setting */
   LONG     MaxHeight;      /* Maximum height setting */
   LONG     LeftLimit;      /* Limits the rendered area from moving too far left */
   LONG     RightLimit;     /* Limits the rendered area from moving too far right */
   LONG     TopLimit;       /* Limits the rendered area from moving too far up */
   LONG     BottomLimit;    /* Limits the rendered area from moving too far down */
   LONG     Frame;          /* Current frame number */
   OBJECTID ScreenID;       /* Object ID of the video display */
   LONG     Flags;          /* Special flags */
   LONG     XCoord;         /* Fixed horizontal coordinate */
   LONG     YCoord;         /* Fixed vertical coordinate */
   LONG     Width;          /* Fixed width */
   LONG     Height;         /* Fixed height */
   OBJECTID RootLayer;      /* Drawable that is acting as a root for many drawable children (useful when applying translucency) */
   OBJECTID TitleObjectID;  /* If the window title changes, the object referred to in this field will have its string set */
   OBJECTID ProgramID;      /* The task that is represented by the render object (important for linking desktop windows to foreign tasks) */
   LONG     Align;          /* Alignment flags */
   LONG     Dimensions;     /* Dimension flags */

   PRIVATE_FIELDS
   struct RGB prvColourRGB;          /* Background colour specification */
   struct Bitmap *prvDrawableBitmap; /* Used as a temporary pointer for the drawing process */
   struct Bitmap *prvBitmap;
   OBJECTID prvBitmapOwnerID;        /* The render object that owns the root bitmap */
   LONG     prvButtonFlags;          /* Reflects the state of any buttons held on the Render object */
   LONG     prvXOffset;              /* Fixed horizontal offset */
   LONG     prvYOffset;              /* Fixed vertical offset */
   FLOAT    prvXOffsetPercent;       /* Relative horizontal offset */
   FLOAT    prvYOffsetPercent;       /* Relative vertical offset */
   FLOAT    prvWidthPercent;         /* Relative width */
   FLOAT    prvHeightPercent;        /* Relative height */
   FLOAT    prvXPercent;             /* Relative horizontal coordinate */
   FLOAT    prvYPercent;             /* Relative vertical coordinate */
   FLOAT    prvAddX;                 /* Horizontal user movement cache */
   FLOAT    prvAddY;                 /* Vertical user movement cache */
   FLOAT    prvAddZ;                 /* Depth user movement cache */
   WINHANDLE prvScreenWindow;        /* Reference to the platform dependent window representing the Render object */
   WORD     prvAnchorX;              /* Horizontal anchor position */
   WORD     prvAnchorY;              /* Vertical anchor position */
   WORD     prvOpacity;
   BYTE     prvBitsPerPixel;         /* Bitmap bits per pixel */
   BYTE     prvBytesPerPixel;        /* Bitmap bytes per pixel */
   LONG     prvLineWidth;            /* Bitmap line width, in bytes */
   MEMORYID prvDataMID;              /* Bitmap memory reference */
   struct RenderCallback *prvCallback;
   struct RenderCallback prvCallbackCache[4];
   UBYTE  prvCallbackCount;
   BYTE   prvPointerWithin;          /* TRUE if the pointer is within the drawable */
   OBJECTID prvAnchorID;
};

/*****************************************************************************
** General Flags
*/

#define RNF_INVISIBLE        0x00000001
#define RNF_STICKTOBACK      0x00000002
#define RNF_STICKTOFRONT     0x00000004
#define RNF_VISIBLE          0x00000008
#define RNF_STICKY           0x00000010
#define RNF_HIDE             0x00000020 /* Private */
#define RNF_CLEAR            0x00000040 /* Read-only */
#define RNF_HASFOCUS         0x00000080
#define RNF_FASTRESIZE       0x00000100 /* Enables fast resize drawing */
#define RNF_DISABLED         0x00000200 /* The render object is disabled */
#define RNF_REGION           0x00000400 /* Private: Indicates a region */
#define RNF_ROOTLAYER        0x00000800
#define RNF_AUTOQUIT         0x00001000
#define RNF_HOST             0x00002000
#define RNF_PRECOPY          0x00004000
#define RNF_VIDEO            0x00008000
#define RNF_NOHORIZONTAL     0x00010000
#define RNF_NOVERTICAL       0x00020000
#define RNF_CURSOR           0x00040000
#define RNF_SCROLLCONTENT    0x00080000
#define RNF_AFTERCOPY        0x00100000
#define RNF_FIXEDBUFFER      0x00200000 /* Passes the NEVERSHRINK option to the render bitmap */
#define RNF_CUSTOMROOT       0x00400000 /* Private */
#define RNF_PERVASIVECOPY    0x00800000
#define RNF_NOFOCUS          0x01000000 /* Prevents any kind of focussing on this object */
#define RNF_FIXEDDEPTH       0x02000000 /* The destinstion buffer always remains at the same depth */

#define RNF_POINTER        (RNF_CURSOR)
#define RNF_PRIVATE        (RNF_HASFOCUS|RNF_CUSTOMROOT|RNF_ROOTLAYER)
#define RNF_WRITEONLY      (RNF_VIDEO)
#define RNF_READONLY       (RNF_HOST|RNF_HIDE|RNF_CLEAR)
#define RNF_VOLATILE       (RNF_PRECOPY|RNF_AFTERCOPY|RNF_CURSOR)

/*****************************************************************************
** Render Methods.
*/

#define MT_InheritedFocus    -1
#define MT_Expose            -2
#define MT_MoveContent       -3
#define MT_InvalidateRegion  -5
#define MT_SetRenderDisplay  -6
#define MT_SetRenderOpacity  -7
#define MT_SetRenderCallback -8

struct mtSetRenderCallback {
   LONG Data;
   APTR Routine;
};

struct mtInvalidateRegion {
   LONG XCoord;
   LONG YCoord;
   LONG Width;
   LONG Height;
};

struct mtExpose {
   LONG XCoord;
   LONG YCoord;
   LONG Width;
   LONG Height;
   LONG Flags;
};

#define EXF_CHILDREN                0x00000001
#define EXF_SIBLINGS                0x00000002
#define EXF_REDRAW_VOLATILE         0x00000004
#define EXF_REDRAW_VOLATILE_OVERLAP 0x00000008

struct mtInheritedFocus {
   OBJECTID FocusID;
};

struct mtMoveContent {
   LONG XCoord;
   LONG YCoord;
   LONG Width;
   LONG Height;
   LONG XDest;
   LONG YDest;
   LONG ClipLeft;
   LONG ClipTop;
   LONG ClipRight;
   LONG ClipBottom;
   LONG Flags;
};

#define MCF_EXPOSE         0x00000001 /* Generate exposure events from source areas that fall outside the clip area */
#define MCF_IGNORECHILDREN 0x00000002 /* Do not try to avoid child areas when moving content */
#define MCF_INVALIDATE     0x00000004 /* Generate an invalidate rather than expose message for outlying areas */

struct mtSetRenderDisplay {
   LONG XCoord;
   LONG YCoord;
   LONG Width;
   LONG Height;
   LONG InsideWidth;
   LONG InsideHeight;
   LONG BitsPerPixel;
   DOUBLE RefreshRate;
   LONG Flags;
};

struct mtSetRenderOpacity {
   DOUBLE Value;
   DOUBLE Adjustment;
};

#endif /* GUI_RENDER_H */
